import numpy as np
import math as mth

from dtg.tail.estimate.hill import HillEstimator


def seq(x, sz=50):
    x_ = HillEstimator.prepare(x)
    ks = HillEstimator.get_k(x)
    g_0 = HillEstimator.estimate(x_, int(2 * mth.sqrt(x.size)))
    r_0 = 0.25 * g_0 * (x.size ** 0.25)
    ga = HillEstimator.estimate(x_, ks)

    mx = np.array(
        [
            np.max([mth.sqrt(i + 1) * (ga[i] - ga[k]) for i in np.arange(1, k)])
            for k in np.arange(2, ga.size)
        ])

    res = []
    for i in np.arange(sz):
        r_1 = r_0 * (sz - i)/sz
        r_2 = r_1 ** 0.7

        k_1 = np.where(mx > r_1)[0]
        if k_1.size == 0:
            continue
        k_1 = k_1[0] + 2
        k_2 = np.where(mx > r_2)[0]
        if k_2.size == 0:
            continue
        k_2 = k_2[0] + 2

        k_op = int(mth.pow(k_2 / mth.pow(k_1, 0.7), 10 / 3) * mth.pow(2 * g_0, 1 / 3) / 3)
        if k_op > x_.size - 2:
            continue
        return HillEstimator.estimate(x_, k_op), k_op
    return None, None



import numpy as np

from dtg.tail.estimate.estimator import TailEstimator


class PickandEstimator(TailEstimator):
    @staticmethod
    def get_k(x):
        return np.arange(1, np.floor(x.size/4)-1)

    @staticmethod
    def estimate(x, k):
        if hasattr(k, '__iter__'):
            return np.array([PickandEstimator.estimate(x, i) for i in k])
        if 4*(k+1) >= x.size:
            k = x.size//4 - 1
        if x[-2*k-1] - x[-4*k-3] == 0:
            return np.NAN
        if (x[-k] - x[-2*k-1])/(x[-2*k-1] - x[-4*k-3]) <= 0:
            return np.NAN
        return np.log((x[-k] - x[-2*k-1])/(x[-2*k-1] - x[-4*k-3]), 2)



from dtg.tail.estimate.hill import HillEstimator
from dtg.tail.estimate.moment import MomentEstimator
from dtg.tail.estimate.pickands import PickandEstimator
from dtg.tail.estimate.ratio import RatioEstimator
from dtg.tail.mse import boot_estimate


def basic_tail(data):
    alp = boot_estimate(HillEstimator, data, 1 / 2, 2 / 3, 100, speed=False)[0]
    print("hill", 1 / alp)

    alp = boot_estimate(RatioEstimator, data, 1 / 2, 2 / 3, 100, speed=False)[0]
    print("ratio", 1 / alp)

    alp = boot_estimate(MomentEstimator, data, 1 / 2, 2 / 3, 100, speed=False)[0]
    print("moment", 1 / alp)

    alp = boot_estimate(PickandEstimator, data, 1 / 2, 2 / 3, 100, speed=False)[0]
    print("pickands", 1 / alp)



import numpy as np

from dtg.tail.estimate.estimator import TailEstimator


class HillEstimator(TailEstimator):
    @staticmethod
    def estimate(x, k):
        if hasattr(k, '__iter__'):
            return np.array([HillEstimator.estimate(x, i) for i in k])
        return np.mean(np.log(x[-k:])) - np.log(x[-k-1])



import numpy as np


def method_higushi(x):
    l = []
    k = np.arange(4, x.size/2)

    for i in k:
        js = []
        for j in np.arange(i):
            sub = x[j:x.size:i]
            c = (x.size-1)/(((j+1)**2) * int((x.size-j-1)/(j+1)))
            js.append(c*np.sum([abs(sub[f+1] - sub[f]) for f in np.arange(sub.size-1)]))
        l.append(np.mean(js))

    return np.array(l), k


def h_higushi(x):
    l, k = method_higushi(x)
    return np.mean(np.divide(np.log(k),np.log(l)))



import numpy as np
import statsmodels as st


def acf(x, h=None):
    if h is None:
        hs = np.arange(2, x.size - 2)
        return acf(x, hs), hs

    if hasattr(h, "__iter__"):
        return np.array([acf(x, i) for i in h])

    mn = np.mean(x)
    return np.sum(np.multiply(x[: x.size - h] - mn, x[h:] - mn)) / np.sum(
        np.power(x - mn, 2)
    )


def acf_inf(x, h=None):
    if h is None:
        hs = np.arange(2, x.size - 2)
        return acf_inf(x, hs), hs

    if hasattr(h, "__iter__"):
        return np.array([acf_inf(x, i) for i in h])

    return np.sum(np.multiply(x[: x.size - h], x[h:])) / np.sum(np.power(x, 2))


def ljungbox(x, h):
    return st.stats.diagnostic.acorr_ljungbox(x)[0:2]



import numpy as np

from scipy.stats import kstest

from dtg.tail.estimate.hill import HillEstimator


def optimal_value(x):
    ks = HillEstimator.get_k(x)
    i = np.sort(x)[::-1]
    x_ = HillEstimator.prepare(x)

    res = []
    exes = []
    for k in ks:
        exes.append(1/HillEstimator.estimate(x_, k))
        fs = lambda x: 1 - x**-exes[-1]
        res.append(kstest(i/i[k], fs, N=1000)[0])

    return exes[np.argmin(res)], ks[np.argmin(res)]



import numpy as np

from scipy.stats import norm


def boot(entity, x, alp, beta, num, k):
    if hasattr(k, "__iter__"):
        if len(k) == 0:
            return [(np.infty, 0, 0)]

        return np.array([boot(entity, x, alp, beta, num, i) for i in k])

    try:
        k1 = int(k * (x.size ** ((1 - beta) * alp)))
        n1 = int(x.size ** beta)
        if k1 + 1 > n1:
            k1 = n1 - 1

        ga = []
        for _ in np.arange(num):
            sub_x = np.array(
                [x[i] for i in np.random.choice(np.arange(0, x.size), n1, replace=True)]
            )
            sub_x.sort()
            ga.append(entity.estimate(sub_x, k1))

        ga = np.array(ga)
        real = entity.estimate(x, k)
        bias = (np.mean(np.where(np.isnan(ga), real-10, ga)) - real) ** 2
        vr = np.var(np.where(np.isnan(ga), np.nanmean(ga), ga))
        return vr + bias, real, np.sqrt(vr)
    except Exception as error:
        print(error)
        return np.infty, 0, 0
    except OverflowError as error:
        print(error)
        return np.infty, 0, 0


def boot_estimate(entity, x, alp, beta, num, speed=True, pers=0.05):
    k = entity.get_k(x, boot=True)
    x_ = entity.prepare(x)

    if len(k) < 2:
        return None, 0, (0, 0)

    if speed:
        count = 0
        mse, alpr, dls = boot(entity, x_, alp, beta, num, k[0])
        k_opt = k[0]

        for k_ in k:
            mse_, alp_, dls_ = boot(entity, x_, alp, beta, num, k_)

            if mse_ < mse:
                mse = mse_
                alpr = alp_
                k_opt = k_
                count = 0
                dls = dls_
            else:
                count += 1

            if count == int(0.1 * k.size) or count == 100:
                break
        return (
            alpr,
            k_opt,
            (alpr - dls * norm.ppf(1 - pers), alpr - dls * norm.ppf(pers)),
        )

    mses, alps, dls = zip(*boot(entity, x_, alp, beta, num, k))
    if len(mses) == 1:
        return mses[0], alps[0], (alps[0], alps[0])
    ar = np.argmin(mses)
    return (
        alps[ar],
        k[ar],
        (
            alps[ar] - dls[ar] * norm.ppf(1 - (pers / 2)),
            alps[ar] - dls[ar] * norm.ppf(pers / 2),
        ),
    )



import numpy as np


class GroupEstimator:
    def __init__(self):
        self.z = 0
        self.l = 0

    def add(self, x):
        k = []
        blk_ = np.sort(x)
        k.append(blk_[-2]/blk_[-1])

        sz = len(k)
        self.z = (self.l*self.z + np.mean(k))/(self.l + sz)
        self.l += sz

    def get_l(self):
        return self.l

    def get_z(self):
        return self.z

    def plus(self, est):
        self.z = self.l*self.z + est.get_l()*est.get_z()
        self.l += est.get_l()
        self.z = self.z/self.l

    def estimate(self):
        return self.z**(-1) - 1



import numpy as np

from dtg.add.transform import symmetrization
from dtg.dependence.acf import acf


def stat(x, h, alp, sym=True, fun=acf):
    n = x.size
    if sym:
        x = symmetrization(x)

    res = (n / np.log(n)) ** (2 / alp) * np.sum(
        [fun(x, i + 1) ** 2 for i in np.arange(h)]
    )
    return res, p_value(res, alp)


def p_value(x, alp):
    return



import numpy as np


def method_rs(x):
    r = np.arange(2, int(x.size/3))

    res = []
    for r_n in r:
        l = []
        for i in np.arange(int(x.size/r_n)):
            sub = x[i*r_n:((i+1)*r_n):1]
            bias = sub - np.mean(sub)
            l.append((np.max(bias)-np.min(bias))/np.mean(np.power(bias,2)))
        res.append(np.mean(l))
    return np.array(res), r


def h_rs(x):
    q, r = method_rs(x)
    return - np.mean(np.divide(np.log(q), np.log(r)))



import itertools

import numpy as np


def stat(data, q=None):
    ns = data.size
    mn = np.mean(data)

    if q is None:
        q = int(np.sqrt(ns))

    dels = data - mn
    s = np.array([np.sum(dels[: i + 1]) for i in np.arange(dels.size)])
    gamma = []
    for _ in np.arange(q):
        for j in np.arange(q):
            gamma.append(np.sum(np.multiply(dels[: ns - j], dels[j:])) / ns)

    v = (np.sum(s ** 2) - (np.sum(s) ** 2 / ns)) / (ns ** 2)
    s = np.sum(gamma) / q
    if s == 0:
        return 0
    return np.abs(v / s)


def undir_stat(data, q=None, mx=None):
    ns = data.size
    mn = np.mean(data)

    if q is None:
        q = int(np.sqrt(ns))
    if mx is None:
        mx = ns

    dels_ = data - mn
    res = []
    for dels in itertools.combinations(dels_, ns):
        dels = np.array(dels)
        s = np.array([np.sum(dels[: i + 1]) for i in np.arange(dels.size)])
        gamma = []
        for _ in np.arange(q):
            for j in np.arange(q):
                gamma.append(np.sum(np.multiply(dels[: ns - j], dels[j:])) / ns)

        v = (np.sum(s ** 2) - (np.sum(s) ** 2 / ns)) / (ns ** 2)
        s = np.sum(gamma) / q
        if s == 0:
            return 0
        res.append(np.abs(v / s))
    return np.mean(res)



import numpy as np


class TailEstimator:
    @staticmethod
    def prepare(x):
        return np.sort(x)

    @staticmethod
    def get_k(x, boot=False):
        if boot:
            return np.arange(5, int(x.size/4))
        return np.arange(2, x.size-2)

    @staticmethod
    def estimate(x, k):
        return None

    @staticmethod
    def get_opt_k(x):
        return None



import numpy as np
import itertools


def stat_exponential(data_split, q):
    sz = np.sum([np.size(data) for data in data_split])
    st = sz * np.log(np.sum([q*np.sum(data) for data in data_split])/sz)
    for data in data_split:
        st -= np.size(data)*np.log(q*np.mean(data))
    return np.sqrt(2*np.abs(st))


def full_stat_exponential(data):
    st0 = np.size(data) * np.log(np.mean(data))
    st = []
    for r in np.arange(1, np.floor((np.size(data)/2))):
        for data_ in itertools.combinations(data, int(r)):
            st.append(st0-r*np.log(np.mean(data_))-(np.size(data_)-r)*np.log((np.sum(data)-np.sum(data_))/(np.size(data_)-r)))
    return np.sqrt(2*np.max(st))



from setuptools import setup, find_packages

version = "1.0.1.5"


setup(
    name="dtg",
    version=version,
    description="Heavy-tailed distribution analysis",
    author="Maksim Ryzhov",
    author_email="ryzhov@phystech.edu",
    license="MIT",
    packages=find_packages(),
    url="https://gitlab.com/BoBeni/DTG",
    install_requires=["numpy", "scipy", "statsmodels"],
    zip_safe=False,
)



import numpy as np

from dtg.tail.estimate.estimator import TailEstimator


class RatioEstimator(TailEstimator):
    @staticmethod
    def estimate(x, k):
        if hasattr(k, '__iter__'):
            return np.array([RatioEstimator.estimate(x, i) for i in k])

        x_n = x[-k-1]
        x_ = x[x > x_n]

        if x_.size == 0:
            return np.NAN

        return np.sum(np.log(x_/x_n))/x_.size



import numpy as np
import math as m

from dtg.tail.estimate.hill import HillEstimator


class MomentEstimator(HillEstimator):
    @staticmethod
    def estimate(x, k):
        if hasattr(k, '__iter__'):
            return np.array([MomentEstimator.estimate(x, i) for i in k])

        h = super(MomentEstimator, MomentEstimator).estimate(x, k)
        s = np.mean([(m.log(x[-i-1]/x[-k-1]))**2 for i in np.arange(k) if x[-i-1]/x[-k-1] > 0])
        if s == 0:
            return np.NAN
        return h + 1 - 0.5*((1 - ((h**2)/s))**2)



import numpy as np

from dtg.tail.estimate.hill import HillEstimator


class UHEstimator(HillEstimator):
    @staticmethod
    def estimate(x, k):
        if hasattr(k, '__iter__'):
            return np.array([UHEstimator.estimate(x, i) for i in k])
        x_ = np.array([x[-i - 1] * (super(UHEstimator, UHEstimator).estimate(x, i)) for i in super(UHEstimator, UHEstimator).get_k(x)])
        if k >= x_.size or k < 0:
            k = x_.size - 1
        return super(UHEstimator, UHEstimator).estimate(x_, k)

    @staticmethod
    def get_k(x):
        res = super(UHEstimator, UHEstimator).get_k(x)
        return np.arange(res.size)
